#include <Halide.h>
#include "halide_generators_common.h"

namespace {

class BoxBlur : public Halide::Generator<BoxBlur> {
public:
    ImageParam input{ Halide::type_of<uint8_t>(), 3, "input" };
    Param<float> scale{ "scale", 1.0 };
    GeneratorParam<int> radius{ "radius", 1 };

    Expr conv_x(Func in, Var x, Var y, Var c, int r, int d) {
        Expr cx = 0.0f;
        for (int i=-r; i<=r; i++) {
            cx += in(x+cast<int>(i*scale), y, c) / d;
        }
        return cx;
    }

    Expr conv_y(Func in, Var x, Var y, Var c, int r, int d) {
        Expr cy = 0.0f;
        for (int i=-r; i<=r; i++) {
            cy += in(x, y+cast<int>(i*scale), c) / d;
        }
        return cy;
    }

    Func conv_xy(Func in, Var x, Var y, Var c, int r, int d) {
        Func cx, cy;
        cx(x, y, c) = conv_x(in, x, y, c, r, d);
        cy(x, y, c) = conv_y(cx, x, y, c, r, d);
        return cy;
    }

    Func build() override {
        Var x("x"), y("y"), c("c");

        Func clamped("clamped");
        Expr clamped_x = clamp(x, 0, input.width()-1);
        Expr clamped_y = clamp(y, 0, input.height()-1);
        clamped(x, y, c) = cast<float>(input(clamped_x, clamped_y, c) / 255.0f);

        Func blur("blur");

        int r = radius;
        int d = radius*2 + 1;
        blur = conv_xy(clamped, x, y, c, r, d);

        // recursive blur
//        blur = conv_xy(blur, x, y, c, r, d);

        // reductions (sum) does not work with GLSL
//        RDom rx(-r, r2);
//        blur_x(x, y, c) = Halide::sum(rx, clamped(x+rx, y, c)) / r2;
//        blur_y(x, y, c) = Halide::sum(rx, blur_x(x, y+rx, c)) / r2;

        // static version for radius == 1 measured to be exactly as performant
//        blur_x(x, y, c) = (clamped(x, y, c) + clamped(x+1, y, c) + clamped(x+2, y, c)) / 3;
//        blur_y(x, y, c) = (blur_x(x, y, c) + blur_x(x, y+1, c) + blur_x(x, y+2, c)) / 3;


        Func function("function");
        function(x, y, c) = cast<uint8_t>(blur(x, y, c) * 255.0f);

        schedule_cpu_and_glsl(input, function, x, y, c);
        set_packed(input);
        set_packed(function.output_buffer());

        return function;
    }

};

Halide::RegisterGenerator<BoxBlur> register_box_blur{"box_blur"};

}
