#include "halide_transform.h"
#include "halide_common.h"
#include <QtCore/QDebug>
#include <QtCore/QMetaObject>
#include <QtCore/QMetaProperty>
#include <QtCore/QElapsedTimer>
#include <QtCore/QStandardPaths>
#include <QtQuick/QQuickWindow>
#include <QtQuick/QQuickItem>
#include <Halide.h>

QHash<Halide::Internal::GeneratorBase*, getOutputGeometry> GeneratorHelperRegistry::outputGeometryFunctions;

HalideTransform::HalideTransform(QObject *parent,
                                 bool resultRequired,
                                 bool customContext) :
    QObject(parent),
    m_allPropertiesValid(false),
    m_renderFbo(0),
    m_textureId(0),
    m_resultRequired(resultRequired),
    m_customContext(customContext),
    m_propertiesSource(this)
{
    // FIXME: make to work in AOT as well (only works with JIT now)
    const Halide::Target &target = Halide::get_jit_target_from_environment();
    if (!target.has_feature(Halide::Target::OpenGL))  {
        m_openGL = false;
        qWarning() << "Running with an OpenGL target will yield better performance. Set HL_JIT_TARGET=host-opengl.";
    } else {
        m_openGL = true;
    }
    findQtQuickGLContext();
}

HalideTransform::~HalideTransform()
{
    destroyOpenGLContext();
}

QString HalideTransform::generatorName() const
{
    return m_generatorName;
}

void HalideTransform::setGeneratorName(QString generatorName)
{
    if (generatorName != m_generatorName) {
        m_generatorName = generatorName;
        updateFunctionFromGenerator();
        matchPropertiesToArguments();
        Q_EMIT generatorNameChanged();
    }
}

QObject* HalideTransform::propertiesSource() const
{
    return m_propertiesSource;
}

void HalideTransform::setPropertiesSource(QObject* propertiesSource)
{
    if (propertiesSource != m_propertiesSource) {
        m_propertiesSource = propertiesSource;
        matchPropertiesToArguments();
        Q_EMIT propertiesSourceChanged();
    }
}

void HalideTransform::updateFunctionFromGenerator()
{
    if (!m_generatorName.isEmpty()) {
        Halide::Internal::GeneratorParamValues params;
        m_generator = Halide::Internal::GeneratorRegistry::create(m_generatorName.toStdString(), params);
        m_function = m_generator->build();
        m_function.compile_jit();
    }
}


void HalideTransform::matchPropertiesToArguments()
{
    if (!m_generator || !m_propertiesSource) {
        return;
    }

    m_signalMappers.clear();
    m_propertiesValid.clear();

    const QMetaObject* meta = m_propertiesSource->metaObject();
    Q_FOREACH(Halide::Argument argument, m_generator->get_filter_arguments()) {
        int propertyIndex = meta->indexOfProperty(argument.name.c_str());
        QString propertyName = QString::fromStdString(argument.name);
        m_propertiesValid[propertyName] = false;
        // FIXME: properties of QQuickItem might conflict with generators arguments
        if (propertyIndex == -1) {
            // FIXME: convert type.code to human readable format
            qWarning() << "Missing property" << propertyName << "of type" << argument.type.code;
        } else {
            QVariant value = m_propertiesSource->property(propertyName.toUtf8().constData());
            m_propertiesValid[propertyName] = setArgumentFromProperty(propertyName, value);
            connectToPropertyChange(propertyName);
        }
    }
    m_allPropertiesValid = !m_propertiesValid.values().contains(false);
}

void HalideTransform::connectToPropertyChange(QString propertyName)
{
    QSharedPointer<QSignalMapper> mapper(new QSignalMapper);
    m_signalMappers.append(mapper);
    mapper->setMapping(m_propertiesSource, propertyName);
    connect(mapper.data(), SIGNAL(mapped(QString)), this, SLOT(onPropertyChanged(QString)));

    const QMetaObject* meta = m_propertiesSource->metaObject();
    int propertyIndex = meta->indexOfProperty(propertyName.toUtf8().constData());
    QMetaProperty prop = meta->property(propertyIndex);
    QMetaMethod propertyChangedSignal = prop.notifySignal();
    int mapIndex = mapper->metaObject()->indexOfSlot("map()");
    QMetaMethod mapSlot = mapper->metaObject()->method(mapIndex);
    connect(m_propertiesSource, propertyChangedSignal, mapper.data(), mapSlot);
}

void HalideTransform::onPropertyChanged(QString propertyName)
{
    QVariant value = m_propertiesSource->property(propertyName.toUtf8().constData());
    m_propertiesValid[propertyName] = setArgumentFromProperty(propertyName, value);
    m_allPropertiesValid = !m_propertiesValid.values().contains(false);
    if (m_allPropertiesValid) {
        scheduleUpdate();
    }
}

bool HalideTransform::setArgumentFromProperty(QString propertyName, QVariant value)
{
    // FIXME: creating parameters should be done ahead of time
    QHash<QString, Halide::Internal::Parameter> parameters;
    Q_FOREACH(Halide::Internal::Parameter parameter, m_generator->get_filter_parameters()) {
        QString name = QString::fromStdString(parameter.name());
        parameters[name] = parameter;
    }

    Halide::Internal::Parameter parameter = parameters[propertyName];
    if (parameter.is_buffer()) {
        QImage image = value.value<QImage>();
        if (image.isNull()) {
            m_images.remove(propertyName);
            return false;
        }

        buffer_t* buf;
        // FIXME: why is this swap needed?
        // Because QImage is BGRA and Halide's generators work with RGBA
        image = image.rgbSwapped();
        buf = halideBufferFromQImage(image);
        buf->host_dirty = true;
        Halide::Buffer buffer(Halide::type_of<uint8_t>(), buf);
        delete buf;
        parameter.set_buffer(buffer);
        m_images[propertyName] = image;
    } else {
        Halide::Type type = parameter.type();
        if (type.is_float()) {
            parameter.set_scalar<float>(value.value<float>());
        } else if (type.is_int()) {
            parameter.set_scalar<int>(value.value<int>());
        } else if (type.is_uint()) {
            parameter.set_scalar<unsigned int>(value.value<unsigned int>());
        } else if (type.is_bool()) {
            parameter.set_scalar<bool>(value.value<bool>());
        }
    }

    return true;
}

bool HalideTransform::apply()
{
    if (!m_allPropertiesValid) {
        qWarning() << "apply(): missing parameters for Halide function.";
        return false;
    }

    if (m_generator) {
        QRect outAndMin;
        getOutputGeometry outAndMinCalculator = GeneratorHelperRegistry::outputGeometryFunctions[m_generator.get()];
        if (outAndMinCalculator) {
            outAndMin = outAndMinCalculator(m_generator.get());
            m_textureSize = QSize(outAndMin.width(), outAndMin.height());
        } else {
            // FIXME: each generator should always register its outAndMin function
            m_textureSize = m_images["input"].size();
            outAndMin = QRect(QPoint(0, 0), m_images["input"].size());
        }

        if (m_openGL) {
            QOpenGLContext* currentContext = QOpenGLContext::currentContext();
            QSurface* currentSurface;
            if (currentContext) {
                currentSurface = currentContext->surface();
            }
            m_context.makeCurrent(&m_surface);

            if (!m_renderFbo || m_renderFbo->size() != m_textureSize) {
                if (m_renderFbo) {
                    delete m_renderFbo;
                    m_renderFbo = 0;
                }
                QOpenGLFramebufferObjectFormat format;
                m_renderFbo = new QOpenGLFramebufferObject(m_textureSize, format);

                buffer_t* output_buf = packedBuffer(NULL, m_textureSize.width(), m_textureSize.height(), 4);
                output_buf->dev = m_renderFbo->texture();
                m_outputBuffer = Halide::Buffer(Halide::type_of<uint8_t>(), output_buf);
                delete output_buf;
            }
            m_renderFbo->bind();

            m_outputBuffer.set_min(outAndMin.x(), outAndMin.y());
            m_function.realize(m_outputBuffer);

            m_renderFbo->release();
            m_textureId = m_renderFbo->texture();
            if (m_resultRequired) {
#if (QT_VERSION >= QT_VERSION_CHECK(5, 4, 0))
                m_result = m_renderFbo->toImage(false);
#else
                m_result = m_renderFbo->toImage().mirrored();
#endif
            }
            m_context.doneCurrent();
            if (currentContext) {
                currentContext->makeCurrent(currentSurface);
            }
        } else {
            m_result = QImage(m_textureSize, QImage::Format_RGB32);
            buffer_t* output_buf = halideBufferFromQImage(m_result);
            Halide::Buffer output_buffer(Halide::type_of<uint8_t>(), output_buf);
            delete output_buf;

            output_buffer.set_min(outAndMin.x(), outAndMin.y());

            m_function.realize(output_buffer);
            m_result = m_result.rgbSwapped();
        }

        return true;
    }

    return false;
}

void HalideTransform::findQtQuickGLContext()
{
    // FIXME: we need the QtQuick's scenegraph context to be
    // current in order to setup context sharing.
    // Relying on QQuickItem's ability to retrieve the context works.
    // Another tested way to achieve the same is to lookup
    // the QQuickWindow via QGuiApplication::topLevelWindows()
    QQuickItem* parentItem = qobject_cast<QQuickItem*>(parent());
    if (parentItem) {
        connect(parentItem, SIGNAL(windowChanged(QQuickWindow*)),
                this, SLOT(onWindowChanged(QQuickWindow*)),
                Qt::UniqueConnection);
        onWindowChanged(parentItem->window());
    }
}

void HalideTransform::createOpenGLContext(QOpenGLContext* context)
{
    if (!m_openGL) {
        return;
    }
    QSurfaceFormat format;
    QOpenGLContext* currentContext = QOpenGLContext::currentContext();
    QSurface* currentContextSurface;
    if (currentContext) {
        currentContextSurface = currentContext->surface();
        // Some GL implementations require that the currently bound context is
        // made non-current before we set up sharing
        currentContext->doneCurrent();
    }
    if (context) {
        format = context->format();
        m_context.setShareContext(context);
    } else {
        format.setDepthBufferSize( 24 );
    }

    m_surface.setFormat(format);
    m_surface.create();
    m_context.setFormat(m_surface.format());
    m_context.create();

    if (currentContext) {
        currentContext->makeCurrent(currentContextSurface);
    }
}

void HalideTransform::destroyOpenGLContext()
{
    if (m_openGL) {
        m_surface.destroy();
        if (m_renderFbo) {
            delete m_renderFbo;
            m_renderFbo = 0;
        }
    }
}

void HalideTransform::scheduleUpdate()
{
    if (m_openGL) {
        doUpdate();
    } else {
        doUpdate();
    }
}

void HalideTransform::doUpdate()
{
    if (m_openGL && !m_context.isValid()) {
        if (m_customContext) {
            return;
        } else {
            findQtQuickGLContext();
            if (!m_context.isValid()) {
                return;
            }
        }
    }

    QElapsedTimer timer;
    timer.start();
    if (apply()) {
        qDebug() << QString("\"%1\" applied in %2 ms (%3x%4)")
                    .arg(m_generatorName)
                    .arg(timer.elapsed())
                    .arg(m_textureSize.width())
                    .arg(m_textureSize.height()).toLocal8Bit().constData();
    }
    Q_EMIT updateCompleted(m_textureId, m_textureSize, m_result);
}

void HalideTransform::onWindowChanged(QQuickWindow* window)
{
    if (window != NULL) {
        QOpenGLContext* context = window->openglContext();
        if (context == NULL) {
            connect(window, SIGNAL(openglContextCreated(QOpenGLContext*)),
                    this, SLOT(onWindowContextCreated(QOpenGLContext*)),
                    Qt::UniqueConnection);
        } else {
            createOpenGLContext(context);
            QMetaObject::invokeMethod(this, "scheduleUpdate", Qt::QueuedConnection);
        }
    } else {
        destroyOpenGLContext();
    }
}

void HalideTransform::onWindowContextCreated(QOpenGLContext* context)
{
    createOpenGLContext(context);
    QMetaObject::invokeMethod(this, "scheduleUpdate", Qt::QueuedConnection);
}

bool HalideTransform::saveToFile(QString fileName)
{
    if (m_result.isNull()) {
        bool resultRequired = m_resultRequired;
        m_resultRequired = true;
        doUpdate();
        m_resultRequired = resultRequired;
    }
    QString directory = QStandardPaths::writableLocation(QStandardPaths::PicturesLocation);
    QString filePath = directory + "/" + fileName;
    bool success = m_result.save(filePath);
    if (!success) {
        qWarning() << "Failed to save image to file:" << filePath;
    }

    return success;
}
