#include "image_transform.h"
#include <QElapsedTimer>
#include <QDebug>
#include <QtConcurrent>
#include <QFuture>
#include <QStandardPaths>

ImageTransform::ImageTransform(QObject *parent) :
    QObject(parent),
    m_hasResult(false)
{
    connect(&m_updateWatcher, SIGNAL(finished()), this, SLOT(updateFinished()));
}

ImageTransform::~ImageTransform()
{
}

QImage ImageTransform::sourceImage() const
{
    return m_sourceImage;
}

QImage ImageTransform::result() const
{
    return m_result;
}

bool ImageTransform::hasResult() const
{
    return m_hasResult;
}

void ImageTransform::setHasResult(bool hasResult)
{
    if (hasResult != m_hasResult) {
        m_hasResult = hasResult;
        Q_EMIT hasResultChanged();
    }
}

void ImageTransform::setSourceImage(QImage sourceImage)
{
    if (sourceImage != m_sourceImage) {
        m_sourceImage = sourceImage;
        scheduleUpdate();
        Q_EMIT sourceImageChanged();
    }
}

bool ImageTransform::saveToFile(QString fileName)
{
    doUpdate();
    QString directory = QStandardPaths::writableLocation(QStandardPaths::PicturesLocation);
    QString filePath = directory + "/" + fileName;
    bool success = m_result.save(filePath);
    if (!success) {
        qWarning() << "Failed to save image to file:" << filePath;
    }

    return success;
}

void ImageTransform::prepare(void)
{
}

QImage ImageTransform::transform(QImage input)
{
    return input;
}

void ImageTransform::doUpdate()
{
    QElapsedTimer timer;
    timer.start();
    if(QThread::currentThread() != this->thread()) {
        QMetaObject::invokeMethod(this, "prepare", Qt::BlockingQueuedConnection);
    } else {
        prepare();
    }
    qDebug() << "Prepare took" << timer.elapsed() << "ms";

    timer.restart();
    m_result = this->transform(m_sourceImage);
    qDebug() << "Transformation took" << timer.elapsed() << "ms";

    Q_EMIT resultChanged();
    setHasResult(!m_result.isNull());
}

void ImageTransform::scheduleUpdate()
{
    if (m_updateWatcher.future().isFinished()) {
        QFuture<void> updateFuture = QtConcurrent::run(this, &ImageTransform::doUpdate);
        m_updateWatcher.setFuture(updateFuture);
    } else {
        m_updatePending = true;
    }
}

void ImageTransform::updateFinished()
{
    if (m_updatePending) {
        m_updatePending = false;
        QFuture<void> updateFuture = QtConcurrent::run(this, &ImageTransform::doUpdate);
        m_updateWatcher.setFuture(updateFuture);
    }
}
